/**************************************************************************
 *
 *	project				 : VisiScript
 *
 *	copyright            : (C) 2002 by Michael Neuroth
 *
 * ------------------------------------------------------------------------
 *
 *  $Source$
 *
 *  $Revision$
 *
 *	$Log$
 *
 ***************************************************************************/
/*********************************************************************************
 *																				 *
 * This file is part of the VisiScript package (a frontend for script languages) *
 *																				 *
 * Copyright (C) 2002 by Michael Neuroth.										 *
 *                                                                               *
 * This program is free software; you can redistribute it and/or modify			 *
 * it under the terms of the GNU General Public License as published by    		 *
 * the Free Software Foundation; either version 2 of the License, or             *
 * (at your option) any later version.                                           *
 *                                                                               *
 ********************************************************************************/

/*
  	Changes:

	23.8.2003   Menu item "Word wrap", "Overwrite" (toggle),
    24.8.2003   Save Script-Extention in config file, Tabspace handling implemented
	30.8.2003   saving the last selected directory implemented
	 1.4.2004   fixes for Zaurus c860 (layout manager), flag for show startup file dialog implemented
	30.4.2004   porting QScintilla back to Qt 2.3.2 for embedded Qt
	 1.5.2004   using QScintilla for the script editor, many new editor features
    27.5.2004   bugfixes, new QScintills version perl lexer using, better configure dialog,	
	            multiple document support, last files list, code folding,

	new features:
	  * syntax hightlighing
	  * find text
	  * replace text
	  * show line numbers
	  * autoindent
	  * tabs as tab or spaces
	  * show whitespaces
	  * Zoom in/out
	  * set ratio of output to script window in config dialog
	  * improve Import Dialog --> use layout manager, needed for c860
	  * bugfix: crash with popen and gnuplot? --> bugfix in buffer-length

	open:
	  (* what about the space between lines? is it configurable in QScintilla?) --> font setting !
	  (* redo/undo icons in der Toolleiste)
	  (* autocompletion)
	  (* add/remove marker, find/jump to marker)
	  (* mehrere Dateien editieren --> tab-view)

	new (11.5.2004):
	  * Bugfix: CheckData() in sltClearScript() called
	  * Bugfix: Open... label changed
	  * Bugfix: Search backwards label changed
	new (14.5.2004):
	  * port to windows
	  * perl parser enabled
	  * restore output ratio after program start
	  * using QScintilla 1.3
	new (17.5.2004):
	  * configure dialog with tabs implemented
	  * code folding implemented (configure dialog and menu item)
	  * configure CR/LF modus
	  * enable undo grouping for replace all
	  * implemented some invisible actions for accelerated keys for the zaurus
	    (CTRL+B and CTRL+E)
	new (19.5.2004):
	  * last read files list implemented
	  * ported to QScintilla 1.3
	new (22.5.2004):
	  * calling refreshProperties() at lexer after language change
	  * working on multi-document support
	  * ok, cancel buttons for configure, find and replace dialog for not zaurus
	  * added accelerator keys for not zaurus
	  * save and load menu items improved for not zaurus
    new (23.5.2004):
	  * accelerators for non zaurus in *.qm files defined
	  * translation handled for non zaurus
    new (24.5.2004):
	  * working on multi-document support
	  * some small bugfixes (setFocus(), better text selection colors)
    new (25.5.2004):
	  * working on multi-document support --> ready !
    new (26.5.2004):
	  * working on multi-document support, patches for Zaurus
    new (27.5.2004):
	  * working on drag and drop support for not Zaurus
	  * split path for tab-name implemented
	  * document changed signal implemented
	new (29.5.2004):
	  * CheckData() splited in CheckActData() and CheckAllData()
	  * Bugfix: filename setting in open was wrong
	  * Bugfix: font update after first show does not work
	  * introduced zvisiscript for better seperation of Zaurus and desktop systems
	new (31.5.2004):
	  * add number for empty documents
*/

#include "visiscript.h"

//#include <iostream>
//using namespace std;

#include "ofontselector.h"
#ifdef _with_opie_filedialog
#include "ofileselector.h"
#include "ofiledialog.h"
#else
#include "filedialogs.h"
#endif

#include "mymledit.h"
#include "codeedit.h"
#include "qextscintillalexercpp.h"
#include "qextscintillalexercsharp.h"
#include "qextscintillalexerpython.h"
#include "qextscintillalexerjava.h"
#include "qextscintillalexerjavascript.h"
#include "qextscintillalexerperl.h"
#include "qextscintillalexeridl.h"
#include "qextscintillalexersql.h"
#include "qextscintillalexerhtml.h"
#include "FontDlg.h"

#include <stdlib.h>
#include <stdio.h>
#include <fstream>
#ifdef ZAURUS
#include <unistd.h>
#include <sys/wait.h>
#else
#include <process.h>
#endif

using namespace std;

#ifdef ZAURUS
#include <qpe/qpeapplication.h>
#include <qpe/qpemenubar.h>
#include <qpe/qpetoolbar.h>
#include <qpe/resource.h>
#include <qpe/fileselector.h>
#include <qpe/config.h>

#define QPERESORUCE(x) Resource::loadPixmap(x),

#define NOT_ZAURUS(x) 0

#define MAX_LAST_FILES	5

#else
#include <qmenubar.h>
#include <qtoolbar.h>
#include <qapplication.h>
#include <qsettings.h>
#include <qfiledialog.h>

#define QPERESORUCE(x)

#define NOT_ZAURUS(x) (x)

#define MAX_LAST_FILES	10

#endif

#include <qaction.h>
#include <qpushbutton.h>
#include <qmultilineedit.h>
//#include <qtextview.h>
#include <qfontdialog.h>
#include <qfont.h>
#include <qfontdatabase.h>
#include <qlineedit.h>
#include <qcombobox.h>
#include <qcheckbox.h>
#include <qspinbox.h>
#include <qradiobutton.h>
#include <qlayout.h>
#include <qwidgetstack.h>
#include <qhbox.h>
#include <qvbox.h>
#include <qdialog.h>
#include <qmessagebox.h>
#include <qclipboard.h>
#include <qtabwidget.h>
#include <qobjectlist.h> 

#include "configuredlgimpl.h"
#include "SaveAsDlg.h"
#include "AboutDlg.h"
#ifdef ZAURUS
#include "FindDlg.h"
#include "ReplaceDlg.h"
#else
#include "finddlgimpl.h"
#include "replacedlgimpl.h"
#endif

#define _TEMP_SCRIPT_FILE		"_tsf.xyz"
#define _TEMP_ERR_FILE			"_err.txt"
#define _TEMP_ERR_FILE_PIPE		"2>_err.txt"
#define _TEMP_OUT_FILE			"_out.txt"
#define _TEMP_OUT_FILE_PIPE		"1>_out.txt"
#define _VISISCRIPT_NAME		"visiscript"

#define _OUTPUT_NAME			"pOutput"
#define _SCRIPT_NAME			"pScript"

const int c_iSmall = 8;
const int c_iMedium = 10;
const int c_iLarge = 12;

const int c_iMinOutput = 80;
const int c_iMaxOutput = 200;
const int c_iMinScript = 80;
const int c_iMaxScript = 40;

const int c_iMaxOutputRatio = 70;
const int c_iMinOutputRatio = 15;

#define LANG_NONE 		1
#define LANG_CPP		2
#define LANG_CSHARP		3
#define LANG_PYTHON		4
#define LANG_JAVA		5
#define LANG_JAVASCRIPT	6
#define LANG_MINSCRIPT	7
#define LANG_RUBY		8
#define LANG_PERL		9
#define LANG_PHP		10
#define LANG_IDL		11
#define LANG_HTML		12
#define LANG_SQL		13
#define LANG_MAX_ID		LANG_SQL+1

// ************************************************************************

int g_iDocCount = 1;

static int u_id = 1;
static int get_unique_id()
{
    return u_id++;
}


static unsigned long GetFileSize( const char *sFileName )
{
	fstream aFile( sFileName, ios::in );

	if( aFile.good() )
	{
		aFile.seekg( 0, ios::end );
		return (unsigned long)aFile.tellg();
	}
	return (unsigned long)-1;		// Datei nicht gefunden !
}

static bool ReadTextFile( const char * sFullFileName, QString & sText )
{
	fstream aFile( sFullFileName, ios::in );
	unsigned long nSize = GetFileSize( sFullFileName );

	if( aFile.good() )
	{
		char * sBuffer = new char [ nSize + 1 ];

		unsigned long i = 0;
		for( i=0; (i<nSize) && (!aFile.eof()); i++ )
		{
			aFile.get( sBuffer[i] );
		}
		sBuffer[ (i==0) ? i : i-1 ] = 0;
		sText = sBuffer;
		delete [] sBuffer;
		return true;
	}
	return false;
}

static bool WriteTextFile( const char * sFullFileName, const char * sText, bool bAppend = false )
{
	fstream aFile( sFullFileName, (bAppend ? ios::app : ios::out) );

	if( aFile.good() )
	{
		aFile << sText;
//cout << "write text " << sFullFileName << endl;
//cout << "text: " << endl << sText << endl;
		return true;
	}
	return false;
}

QString GetExtentionFromScriptCmd( const QString & sCmd )
{
	QString sRet;

	if( sCmd == "minscript" )
	{
		sRet = "text/minscript";
	}
	else if( sCmd == "python" )
	{
		sRet = "text/python";
	}
	else if( sCmd == "perl" )
	{
		sRet = "text/perl";
	}
	else if( sCmd == "ruby" )
	{
		sRet = "text/ruby";
	}
	else
	{
		sRet = "text/plain";
	}

	return sRet;
}

// ************************************************************************
/*
 *  Constructs a VisiScript which is a child of 'parent', with the
 *  name 'name' and widget flags set to 'f'
 */
VisiScript::VisiScript( QWidget* parent,  const char* name, WFlags fl )
    : QMainWindow( parent, name, fl ),
	  m_pFileSelector( 0 ),
	  m_pTabWidget( 0 ),
	  m_sScriptCmd( "minscript" ),
	  m_sScriptExtention( "text/plain" ),
	  m_sFontName( "fixed" ),
	  m_sFontStyle( "normal" ),
	  m_iFontSize( c_iMedium ),
	  m_bSaveOutput( false ),
	  m_iWordWrap( 0 ),
	  m_iOverwrite( 0 ),
	  m_iSpacesForTabs( 0 ),
	  m_iShowWhitespaces( 0 ),
	  m_iShowLineNumbers( 1 ),
	  m_iAutoIndent( 0 ),
	  m_iCodeFolding( 0 ),
	  m_iTabSpace( 4 ),
	  m_iLanguageID( LANG_MINSCRIPT ),
	  m_iEolMode( QextScintilla::EolUnix ),
	  m_sExeclCmd( "/bin/bash" ),
	  m_iExecuteMode( EXECUTE_MODE_POPEN ),
	  m_iShowStartupFileDialog( 1 ),
	  m_iOutputRatio( 30 ),
	  m_bIsReplace( false ),
	  m_bRegularExpr( false ),
	  m_bCaseSensitive( false ),
	  m_bOnlyWholeWords( false ),
	  m_bSearchBackward( false ),
	  m_bReplaceAll( false )
{
	loadInit();

	m_aLastFilesList.setAutoDelete( TRUE );

    setToolBarsMovable( FALSE );

	// *************************************************************
	// *************************************************************

#ifdef ZAURUS
    m_pMenu = new QPEToolBar( this );
    m_pMenuBar = new QPEMenuBar( m_pMenu );
    m_pButtonBar = new QPEToolBar( this );
#else
    m_pMenu = new QToolBar( this );
    m_pMenuBar = new QMenuBar( m_pMenu );
    m_pButtonBar = new QToolBar( this );
	setAcceptDrops( TRUE );
#endif

    m_pMenu->setHorizontalStretchable( TRUE );

    pFile = new QPopupMenu( this );
    pEdit = new QPopupMenu( this );
    pView = new QPopupMenu( this );
	pLanguages = new QPopupMenu( pView );

    connect( pFile, SIGNAL( aboutToShow() ), this, SLOT( sltFileAboutToShow() ) );
    connect( pEdit, SIGNAL( aboutToShow() ), this, SLOT( sltEditAboutToShow() ) );
    connect( pView, SIGNAL( aboutToShow() ), this, SLOT( sltViewAboutToShow() ) );
    connect( pLanguages, SIGNAL( aboutToShow() ), this, SLOT( sltLanguagesAboutToShow() ) );
    connect( pLanguages, SIGNAL( activated(int) ), this, SLOT( sltLanguageActivated(int) ) );

    m_pMenuBar->insertItem( tr( "File" ), pFile );
    m_pMenuBar->insertItem( tr( "Edit" ), pEdit );
    m_pMenuBar->insertItem( tr( "View" ), pView );

	// ++++++++++++++++++++++++++++++++++++++++++

    QAction * aAction = new QAction( tr( "New" ), QPERESORUCE( "new" ) QString::null, NOT_ZAURUS( CTRL+Key_N ), this, 0 );
    connect( aAction, SIGNAL( activated() ), this, SLOT( sltFileNew() ) );
    //aAction->addTo( m_pButtonBar );
    aAction->addTo( pFile );

    aAction = new QAction( tr( "Clear" ), QString::null, NOT_ZAURUS( ALT+Key_C ), this, 0 );
    connect( aAction, SIGNAL( activated() ), this, SLOT( sltClearActDocument() ) );
    aAction->addTo( pFile );

    aAction = new QAction( tr( "Close" ), QString::null, NOT_ZAURUS( ALT+Key_X ), this, 0 );
    connect( aAction, SIGNAL( activated() ), this, SLOT( sltCloseActDocument() ) );
    aAction->addTo( pFile );

	pFile->insertSeparator();

    aAction = new QAction( tr( "Open..." ), QPERESORUCE( "fileopen" ) QString::null, NOT_ZAURUS( CTRL+Key_O ), this, 0 );
    connect( aAction, SIGNAL( activated() ), this, SLOT( sltFileOpen() ) );
    //aAction->addTo( m_pButtonBar );
    aAction->addTo( pFile );

    aAction = new QAction( tr( "Save" ), /*Resource::loadPixmap( "filesave" ),*/ QString::null, NOT_ZAURUS( CTRL+Key_S ), this, 0 );
    connect( aAction, SIGNAL( activated() ), this, SLOT( sltFileSave() ) );
    aAction->addTo( pFile );

    aAction = new QAction( tr( "Save as..." ), QString::null, 0, this, 0 );
    connect( aAction, SIGNAL( activated() ), this, SLOT( sltFileSaveAs() ) );
    aAction->addTo( pFile );

	pFile->insertSeparator();

    aAction = new QAction( tr( "Import..." ), QString::null, NOT_ZAURUS( CTRL+Key_I ), this, 0 );
    connect( aAction, SIGNAL( activated() ), this, SLOT( sltFileImport() ) );
    aAction->addTo( pFile );

    aAction = new QAction( tr( "Export..." ), QString::null, 0, this, 0 );
    connect( aAction, SIGNAL( activated() ), this, SLOT( sltFileExport() ) );
    aAction->addTo( pFile );

    aAction = new QAction( tr( "Export output..." ), QString::null, 0, this, 0 );
    connect( aAction, SIGNAL( activated() ), this, SLOT( sltFileSaveOutputAs() ) );
    aAction->addTo( pFile );

	pLastOpendFiles = new QPopupMenu( pFile );
	connect( pLastOpendFiles, SIGNAL( aboutToShow() ), this, SLOT( sltUpdateLastFilesMenu() ) );
	connect( pLastOpendFiles, SIGNAL( activated(int) ), this, SLOT( sltLastFilesMenuActivated(int) ) );
	pFile->insertItem( tr( "Last files" ), pLastOpendFiles );

	pFile->insertSeparator();

	// "settings"
    aAction = new QAction( tr( "About..." ), QString::null, 0, this, 0 );
    connect( aAction, SIGNAL( activated() ), this, SLOT( sltShowAbout() ) );
    aAction->addTo( pFile );

	// ++++++++++++++++++++++++++++++++++++++++++

    aAction = new QAction( tr( "Undo" ), QString::null, NOT_ZAURUS( CTRL+Key_Z ), this, 0 );
    connect( aAction, SIGNAL( activated() ), this, SLOT( sltEditUndo() ) );
    aAction->addTo( pEdit );

    aAction = new QAction( tr( "Redo" ), QString::null, NOT_ZAURUS( CTRL+Key_Y ), this, 0 );
    connect( aAction, SIGNAL( activated() ), this, SLOT( sltEditRedo() ) );
    aAction->addTo( pEdit );

    pEdit->insertSeparator();

	m_pCutAction = new QAction( tr( "Cut" ), QString::null, NOT_ZAURUS( CTRL+Key_X ), this, 0 );
    connect( m_pCutAction, SIGNAL( activated() ), this, SLOT( sltEditCut() ) );
    m_pCutAction->addTo( pEdit );

    m_pCopyAction = new QAction( tr( "Copy" ), QString::null, NOT_ZAURUS( CTRL+Key_C ), this, 0 );
    connect( m_pCopyAction, SIGNAL( activated() ), this, SLOT( sltEditCopy() ) );
    m_pCopyAction->addTo( pEdit );

    m_pPasteAction = new QAction( tr( "Paste" ), QString::null, NOT_ZAURUS( CTRL+Key_V ), this, 0 );
    connect( m_pPasteAction, SIGNAL( activated() ), this, SLOT( sltEditPaste() ) );
    m_pPasteAction->addTo( pEdit );

    pEdit->insertSeparator();

    aAction = new QAction( tr( "Find..." ), QPERESORUCE( "find" ) QString::null, NOT_ZAURUS( CTRL+Key_F ), this, 0 );
    connect( aAction, SIGNAL( activated() ), this, SLOT( sltEditFindText() ) );
    aAction->addTo( m_pButtonBar );
    aAction->addTo( pEdit );

    aAction = new QAction( tr( "Find next" ), QPERESORUCE( "forward" ) QString::null, NOT_ZAURUS( Key_F3 ), this, 0 );
    connect( aAction, SIGNAL( activated() ), this, SLOT( sltEditFindNext() ) );
    aAction->addTo( m_pButtonBar );
    aAction->addTo( pEdit );

    aAction = new QAction( tr( "Replace..." ), QString::null, NOT_ZAURUS( CTRL+Key_H ), this, 0 );
    connect( aAction, SIGNAL( activated() ), this, SLOT( sltEditReplaceText() ) );
    aAction->addTo( pEdit );

    pEdit->insertSeparator();

    aAction = new QAction( tr( "Clear script" ), QString::null, 0, this, 0 );
    connect( aAction, SIGNAL( activated() ), this, SLOT( sltClearScript() ) );
    aAction->addTo( pEdit );

    aAction = new QAction( tr( "Clear output" ), QPERESORUCE( "editdelete" ) QString::null, 0, this, 0 );
    connect( aAction, SIGNAL( activated() ), this, SLOT( sltClearOutput() ) );
    aAction->addTo( m_pButtonBar );
    aAction->addTo( pEdit );

	// ++++++++++++++++++++++++++++++++++++++++++

    pEdit->insertSeparator();

    aAction = new QAction( tr( "Run" ), QPERESORUCE( /*"exec"*/"play" ) QString::null, NOT_ZAURUS( Key_F5 ), this, 0 );
    connect( aAction, SIGNAL( activated() ), this, SLOT( sltRun() ) );
    aAction->addTo( m_pButtonBar );
    aAction->addTo( pEdit );

	// ++++++++++++++++++++++++++++++++++++++++++
/*
	m_pFixedAction = new QAction( tr( "Fixed" ), QString::null, 0, this, 0, TRUE );
    connect( m_pFixedAction, SIGNAL( activated() ), this, SLOT( sltSelectFixed() ) );
    m_pFixedAction->addTo( pView );
	m_pHelveticaAction = new QAction( tr( "Helvetica" ), QString::null, 0, this, 0, TRUE );
    connect( m_pHelveticaAction, SIGNAL( activated() ), this, SLOT( sltSelectHelvetica() ) );
    m_pHelveticaAction->addTo( pView );
	m_pUnifontAction = new QAction( tr( "Unifont" ), QString::null, 0, this, 0, TRUE );
    connect( m_pUnifontAction, SIGNAL( activated() ), this, SLOT( sltSelectUnifont() ) );
    m_pUnifontAction->addTo( pView );

    pView->insertSeparator();

	m_pLargeAction = new QAction( tr( "Large" ), QString::null, 0, this, 0, TRUE );
    connect( m_pLargeAction, SIGNAL( activated() ), this, SLOT( sltSelectLarge() ) );
    m_pLargeAction->addTo( pView );
	m_pMediumAction = new QAction( tr( "Medium" ), QString::null, 0, this, 0, TRUE );
    connect( m_pMediumAction, SIGNAL( activated() ), this, SLOT( sltSelectMedium() ) );
    m_pMediumAction->addTo( pView );
	m_pSmallAction = new QAction( tr( "Small" ), QString::null, 0, this, 0, TRUE );
    connect( m_pSmallAction, SIGNAL( activated() ), this, SLOT( sltSelectSmall() ) );
    m_pSmallAction->addTo( pView );
*/
    aAction = new QAction( tr( "Font..." ), QString::null, 0, this, 0 );
    connect( aAction, SIGNAL( activated() ), this, SLOT( sltSelectFont() ) );
    aAction->addTo( pView );
/*
    pView->insertSeparator();

	aAction = new QAction( tr( "Zoom in" ), QString::null, 0, this, 0, FALSE );
    connect( aAction, SIGNAL( activated() ), this, SLOT( sltZoomIn() ) );
    aAction->addTo( pView );
	aAction = new QAction( tr( "Zoom out" ), QString::null, 0, this, 0, FALSE );
    connect( aAction, SIGNAL( activated() ), this, SLOT( sltZoomOut() ) );
    aAction->addTo( pView );
*/
	// "settings"
    aAction = new QAction( tr( "Configuration..." ), QString::null, 0, this, 0 );
    connect( aAction, SIGNAL( activated() ), this, SLOT( sltConfiguration() ) );
    aAction->addTo( pView );

	pView->insertItem( tr("Language"), pLanguages );
	pLanguages->insertItem( tr("none"), LANG_NONE );
	pLanguages->insertItem( tr("minscript"), LANG_MINSCRIPT );
	pLanguages->insertItem( tr("C++"), LANG_CPP );
	pLanguages->insertItem( tr("C#"), LANG_CSHARP );
	pLanguages->insertItem( tr("Python"), LANG_PYTHON );
	pLanguages->insertItem( tr("Java"), LANG_JAVA );
	pLanguages->insertItem( tr("JavaScript"), LANG_JAVASCRIPT );
		pLanguages->insertItem( tr("Ruby"), LANG_RUBY );
	pLanguages->insertItem( tr("Perl"), LANG_PERL );
		pLanguages->insertItem( tr("PHP"), LANG_PHP );
	pLanguages->insertItem( tr("IDL"), LANG_IDL );
	pLanguages->insertItem( tr("HTML"), LANG_HTML );
	pLanguages->insertItem( tr("SQL"), LANG_SQL );

	pView->insertSeparator();

    aAction = new QAction( tr( "Zoom in" ), QString::null, 0, this, 0 );
    connect( aAction, SIGNAL( activated() ), this, SLOT( sltZoomIn() ) );
    aAction->addTo( pView );
    aAction = new QAction( tr( "Zoom out" ), QString::null, 0, this, 0 );
    connect( aAction, SIGNAL( activated() ), this, SLOT( sltZoomOut() ) );
    aAction->addTo( pView );

	pView->insertSeparator();

	m_pMaximizeScript = new QAction( tr( "Maximize script" ), QString::null, 0, this, 0, TRUE );
    connect( m_pMaximizeScript, SIGNAL( activated() ), this, SLOT( sltMaximizeScript() ) );
    m_pMaximizeScript->addTo( pView );
    m_pMaximizeOutput = new QAction( tr( "Maximize output" ), QString::null, 0, this, 0, TRUE );
    connect( m_pMaximizeOutput, SIGNAL( activated() ), this, SLOT( sltMaximizeOutput() ) );
    m_pMaximizeOutput->addTo( pView );

	pView->insertSeparator();

	m_pToggleWordWrap = new QAction( tr( "Word wrap" ), QString::null, 0, this, 0, TRUE );
    connect( m_pToggleWordWrap, SIGNAL( activated() ), this, SLOT( sltToggleWordWrap() ) );
    m_pToggleWordWrap->addTo( pView );
    m_pToggleOverwrite = new QAction( tr( "Overwrite" ), QString::null, 0, this, 0, TRUE );
    connect( m_pToggleOverwrite, SIGNAL( activated() ), this, SLOT( sltToggleOverwrite() ) );
    m_pToggleOverwrite->addTo( pView );
    m_pCodeFolding = new QAction( tr( "Code folding" ), QString::null, 0, this, 0, TRUE );
    connect( m_pCodeFolding, SIGNAL( activated() ), this, SLOT( sltCodeFolding() ) );
    m_pCodeFolding->addTo( pView );

//#ifdef ZAURUS
//	m_pGotoBegin = new QAction( tr( "Begin" ), QString::null, CTRL+Key_B, this, 0, TRUE );
//	connect( m_pGotoBegin, SIGNAL( activated() ), this, SLOT( sltGotoBegin() ) );
//	m_pGotoEnd = new QAction( tr( "End" ), QString::null, CTRL+Key_E, this, 0, TRUE );
//	connect( m_pGotoEnd, SIGNAL( activated() ), this, SLOT( sltGotoEnd() ) );
//#endif

	// *************************************************************
	// *************************************************************

	m_pStack = new QWidgetStack( this );
    setCentralWidget( m_pStack );

    m_pTabWidget = new QTabWidget( m_pStack, "pTabWidget" );

	m_pStack->addWidget( m_pTabWidget, get_unique_id() );

#ifndef ZAURUS
	AddDocument( g_iDocCount++ );

	updateFileSelector();

	if( m_iShowStartupFileDialog )
	{
	    sltFileOpen();
	}
	else
	{
		sltShowScript();
		updateCaption();
	}

	resize( 800, 600 );
	move( 5, 5 );
#endif

#ifndef _with_opie_filedialog
	m_pExportFileDlg = new FileDialogS(this, "save",TRUE, WType_Modal, TRUE);
	connect( m_pExportFileDlg, SIGNAL( doneLoad() ), this, SLOT( sltDoImport() ) );
    connect( m_pExportFileDlg, SIGNAL( done() ), this, SLOT( sltDoExport() ) );
	m_pExportFileDlg->adjustSize();
#endif
}

/*
 *  Destroys the object and frees any allocated resources
 */
VisiScript::~VisiScript()
{
#ifndef _with_opie_filedialog
	delete m_pExportFileDlg;
#endif
	// no need to delete child widgets, Qt does it all for us
}

void VisiScript::sltFileNew()
{
	AddDocument( g_iDocCount++ );
}

void VisiScript::updateFileSelector()
{
}

void VisiScript::sltFileOpen()
{
#ifndef ZAURUS
	QString sFileName = QFileDialog::getOpenFileName( ".", QString::null, this );

	if( !sFileName.isEmpty() )
	{
		DoImportFile( sFileName );
	}
#endif
}

void VisiScript::sltFileSave()
{
	DoExportFile( GetActScript()->fileName() );
}

void VisiScript::sltFileSaveAs()
{
#ifndef ZAURUS
	QString sFileName = QFileDialog::getSaveFileName( ".", QString::null, this );

	if( !sFileName.isEmpty() )
	{
		GetActScript()->setFileName( sFileName );

		if( DoExportFile( GetActScript()->fileName() ) )
		{
			PushLastFileName( sFileName );
		}
	}
#endif
}

void VisiScript::sltFileImport()
{
#ifdef _with_opie_filedialog
	// NOT USED JET !!!

	//CheckData();

	QString sFileName = OFileDialog::getOpenFileName( OFileSelector::NORMAL );

	DoImportFile( sFileName );
#else
	m_pExportFileDlg->setLoad( TRUE );
	m_pExportFileDlg->setDirectory( m_sLastDirectory );
	m_pExportFileDlg->exec();
	m_pExportFileDlg->adjustSize();
#endif
}

void VisiScript::sltFileExport()
{
#ifdef _with_opie_filedialog
	// NOT USED JET !!!
	
	QString sFileName = OFileDialog::getSaveFileName( OFileSelector::NORMAL );
	//cout << "Export: " << (const char *)sFileName << endl;

	if( !sFileName.isNull() )
	{
		DoExportFile( sFileName );
	}
#else
	m_bSaveOutput = false;
	m_pExportFileDlg->setLoad( FALSE );
	m_pExportFileDlg->setDirectory( m_sLastDirectory );
	m_pExportFileDlg->exec();
	m_pExportFileDlg->adjustSize();
#endif
}

void VisiScript::sltFileSaveOutputAs()
{
#ifdef _with_opie_filedialog
	// NOT USED JET !!!

	QString sFileName = OFileDialog::getSaveFileName( OFileSelector::NORMAL );
	//cout << "SaveOutput: " << (const char *)sFileName << endl;

	if( !sFileName.isNull() )
	{
		QString sText = GetActOutput()->text();
		/*bool bOk =*/ WriteTextFile( sFileName.latin1(), sText );
	}
#else
	m_bSaveOutput = true;
	m_pExportFileDlg->setLoad( FALSE );
	m_pExportFileDlg->setDirectory( m_sLastDirectory );
	m_pExportFileDlg->exec();
	m_pExportFileDlg->adjustSize();
#endif
}

#ifndef _with_opie_filedialog
void VisiScript::sltDoImport()
{
	//CheckData();

	QString * pFileName = m_pExportFileDlg->getAbsSelectedFileName();

	m_sLastDirectory = m_pExportFileDlg->getDirectory();

	DoImportFile( *pFileName );

	delete pFileName;
}

void VisiScript::sltDoExport()
{
	QString * pFileName = m_pExportFileDlg->getAbsSelectedFileName();

	m_sLastDirectory = m_pExportFileDlg->getDirectory();

	DoExportFile( *pFileName );

	delete pFileName;
}
#endif

static void SetComboBoxToText( QComboBox * pCombo, const QString & sText )
{
	for( int i=0; i<pCombo->count(); i++ )
	{
		if( pCombo->text( i ) == sText )
		{
			pCombo->setCurrentItem( i );
			break;
		}
	}
}

void VisiScript::sltConfiguration()
{
	ConfigureDlgImpl * pDlg = new ConfigureDlgImpl( this, "configuredlg", /*bModal*/TRUE );

	pDlg->adjustSize();

	pDlg->m_pScriptCmd->setText( m_sScriptCmd );

	SetComboBoxToText( pDlg->m_pScriptCmdList, m_sScriptCmd );
	SetComboBoxToText( pDlg->m_pType, m_sScriptExtention );

	pDlg->m_pTabSpace->setText( QString::number( m_iTabSpace ) );
	pDlg->m_pSpacesForTabs->setChecked( (bool)m_iSpacesForTabs );
	pDlg->m_pShowWhitespaces->setChecked( (bool)m_iShowWhitespaces );
	pDlg->m_pShowLineNumbers->setChecked( (bool)m_iShowLineNumbers );
	pDlg->m_pAutoIndent->setChecked( (bool)m_iAutoIndent );
	pDlg->m_pCodeFolding->setChecked( (bool)m_iCodeFolding );
	pDlg->m_pOutputRatio->setValue( m_iOutputRatio );

	pDlg->m_pExecuteCmd->setText( m_sExeclCmd );

	if( m_iExecuteMode == EXECUTE_MODE_POPEN )
	{
		pDlg->m_pUsepopen->setChecked( TRUE );
	}
	if( m_iExecuteMode == EXECUTE_MODE_EXECL )
	{
		pDlg->m_pUseexecl->setChecked( TRUE );
	}
	if( m_iShowStartupFileDialog==0 )
	{
		pDlg->m_pShowFileDialog->setChecked( FALSE );
	}
	else
	{
		pDlg->m_pShowFileDialog->setChecked( TRUE );
	}

	if( m_iEolMode==QextScintilla::EolWindows )
	{
		pDlg->m_pCRLFModus->setCurrentItem( 0 );
	}
	else if( m_iEolMode==QextScintilla::EolUnix )
	{
		pDlg->m_pCRLFModus->setCurrentItem( 1 );
	}
	else if( m_iEolMode==QextScintilla::EolMac )
	{
		pDlg->m_pCRLFModus->setCurrentItem( 2 );
	}

	pDlg->adjustSize();

	int iRet = pDlg->exec();

	if( iRet == 1 )
	{
		// ok, pressed
		m_sScriptCmd = pDlg->m_pScriptCmd->text();

		m_sScriptExtention = pDlg->m_pType->currentText();

		m_iSpacesForTabs = (int)pDlg->m_pSpacesForTabs->isChecked();
		m_iShowWhitespaces = (int)pDlg->m_pShowWhitespaces->isChecked();
		m_iShowLineNumbers = (int)pDlg->m_pShowLineNumbers->isChecked();
		m_iAutoIndent = (int)pDlg->m_pAutoIndent->isChecked();
		m_iCodeFolding = (int)pDlg->m_pCodeFolding->isChecked();
		m_iOutputRatio = pDlg->m_pOutputRatio->value();
		int iMode = pDlg->m_pCRLFModus->currentItem();
		if( iMode == 0 )
		{
			m_iEolMode = QextScintilla::EolWindows;
		}
		else if( iMode == 1 )
		{
			m_iEolMode = QextScintilla::EolUnix;
		}
		else if( iMode == 2 )
		{
			m_iEolMode = QextScintilla::EolMac;
		}

		m_iTabSpace = pDlg->m_pTabSpace->text().toInt();
#ifdef ZAURUS
		QMultiLineEdit::setDefaultTabStop( m_iTabSpace );
#else
		ForAllDocumentOutput( m_iTabSpace, MyQMultiLineEdit::setTabStopWidth );
#endif
		ForAllDocumentScripts( m_iTabSpace, &MyQCodeEdit::setTabWidth );
		ForAllDocumentScripts( true, (void (MyQCodeEdit::*)(bool))&MyQCodeEdit::repaint );

		m_sExeclCmd = pDlg->m_pExecuteCmd->text();

		if( pDlg->m_pUsepopen->isChecked() )
		{
			m_iExecuteMode = EXECUTE_MODE_POPEN;
		}
		if( pDlg->m_pUseexecl->isChecked() )
		{
			m_iExecuteMode = EXECUTE_MODE_EXECL;
		}
		if( pDlg->m_pShowFileDialog->isChecked() )
		{
			m_iShowStartupFileDialog = 1;
		}
		else
		{
			m_iShowStartupFileDialog = 0;
		}
		
		UpdateFlags();
		UpdateScriptOutputRatio();

		saveInit();
	}
	delete pDlg;
}

void VisiScript::sltShowAbout()
{
	AboutDlg * pDlg = new AboutDlg( this, "aboutdlg", /*bModal*/TRUE );

	pDlg->adjustSize();

	/*int iRet =*/ pDlg->exec();

	delete pDlg;
}

void VisiScript::sltUpdateLastFilesMenu()
{
	pLastOpendFiles->clear();

	for( unsigned int i=0; i<m_aLastFilesList.count(); i++ )
	{
		QString sLastFileName = *(m_aLastFilesList.at( m_aLastFilesList.count()-i-1 ));

		// ** show filename with number 

#ifndef ZAURUS
		sLastFileName = QString( "&" ) + QString::number( i ) + QString( " " ) + 
						sLastFileName;
#endif

		pLastOpendFiles->insertItem( sLastFileName, i );
	}
}

void VisiScript::sltLastFilesMenuActivated( int iIndex )
{
	QString sFileName = pLastOpendFiles->text( iIndex );

#ifndef ZAURUS
	sFileName = sFileName.right( sFileName.length()-3 );
#endif

	DoImportFile( sFileName );
}

void VisiScript::sltUpdateChilds()
{
	UpdateFont();
}

void VisiScript::sltSelectFont()
{
#ifndef QT_NO_FONTDIALOG
	bool bOk = false;
	QFont aFont( m_sFontName, m_iFontSize );

	aFont = QFontDialog::getFont( &bOk, aFont, this );
	if( bOk )
	{
		m_sFontName = aFont.family();
		m_sFontStyle = aFont.styleHint();
		m_iFontSize = aFont.pointSize();
		UpdateFont();
		saveInit();
	}
#else
	FontDlg * pDlg = new FontDlg( this, "fontdlg", /*bModal*/TRUE );

	pDlg->m_pFontWidget->setSelectedFont( m_sFontName, m_sFontStyle, m_iFontSize );

	pDlg->adjustSize();

	int iRet = pDlg->exec();

	if( iRet == 1 )
	{
		m_sFontName = pDlg->m_pFontWidget->fontFamily();
		m_sFontStyle = pDlg->m_pFontWidget->fontStyle();
		m_iFontSize = pDlg->m_pFontWidget->fontSize();
		UpdateFont();
		saveInit();
	}

	delete pDlg;
#endif
}
/*
void VisiScript::sltSelectHelvetica()
{
	m_pFixedAction->setOn( false );
	m_pUnifontAction->setOn( false );
	m_sFontName = "Helvetica";
	UpdateFont();
	saveInit();
}

void VisiScript::sltSelectFixed()
{
	m_pHelveticaAction->setOn( false );
	m_pUnifontAction->setOn( false );
	m_sFontName = "fixed";
	UpdateFont();
	saveInit();
}

void VisiScript::sltSelectUnifont()
{
	m_pFixedAction->setOn( false );
	m_pHelveticaAction->setOn( false );
	m_sFontName = "Unifont";
	UpdateFont();
	saveInit();
}

void VisiScript::sltSelectSmall()
{
	m_pMediumAction->setOn( false );
	m_pLargeAction->setOn( false );
	m_iFontSize = c_iSmall;
	UpdateFont();
	saveInit();
}

void VisiScript::sltSelectMedium()
{
	m_pSmallAction->setOn( false );
	m_pLargeAction->setOn( false );
	m_iFontSize = c_iMedium;
	UpdateFont();
	saveInit();
}

void VisiScript::sltSelectLarge()
{
	m_pMediumAction->setOn( false );
	m_pSmallAction->setOn( false );
	m_iFontSize = c_iLarge;
	UpdateFont();
	saveInit();
}
*/

void VisiScript::sltZoomIn()
{
	m_iFontSize++;
	UpdateFont();
	saveInit();
}

void VisiScript::sltZoomOut()
{
	m_iFontSize--;
	UpdateFont();
	saveInit();
}

void VisiScript::sltMaximizeScript()
{
	if( !m_pMaximizeScript->isOn() )
	{
		SetOutputSize( m_iOutputRatio );
	}
	else
	{
		m_pMaximizeOutput->setOn( false );
		SetOutputSize( c_iMinOutputRatio );
	}
}

void VisiScript::sltMaximizeOutput()
{
	if( !m_pMaximizeOutput->isOn() )
	{
		SetOutputSize( m_iOutputRatio );
	}
	else
	{
		m_pMaximizeScript->setOn( false );
		SetOutputSize( c_iMaxOutputRatio );
	}
}

void VisiScript::sltToggleWordWrap()
{
	if( !m_pToggleWordWrap->isOn() )
	{
		ForAllDocumentScripts( false, &MyQCodeEdit::setWordWrap );
		m_iWordWrap = 0;
	}
	else
	{
		ForAllDocumentScripts( true, &MyQCodeEdit::setWordWrap );
		m_iWordWrap = 1;
	}
}

void VisiScript::sltToggleOverwrite()
{
	if( !m_pToggleOverwrite->isOn() )
	{
		ForAllDocumentScripts( false, &MyQCodeEdit::setOverwriteMode );
	}
	else
	{
		ForAllDocumentScripts( true, &MyQCodeEdit::setOverwriteMode );
	}
}

void VisiScript::sltCodeFolding()
{
	GetActScript()->foldAll();
}

void VisiScript::UpdateFont()
{
	QFontDatabase aFontDB;
	int iWeight = aFontDB.weight( m_sFontName, m_sFontStyle );
	bool bItalic = aFontDB.italic( m_sFontName, m_sFontStyle );
	QFont aTempFont( m_sFontName, m_iFontSize, iWeight, bItalic );

	for( int i=0; i<GetDocumentCount(); i++ )
	{
		GetScriptFor( i )->setFont( aTempFont );
		GetScriptFor( i )->updateFontForLexer( aTempFont );

		GetOutputFor( i )->setFont( aTempFont );
	}
}

void VisiScript::UpdateFlags()
{
	// restore values form the init/configuration file
	if( m_iWordWrap==0 )
	{
		m_pToggleWordWrap->setOn( false );
		ForAllDocumentScripts( false, &MyQCodeEdit::setWordWrap );
	}
	else
	{
		m_pToggleWordWrap->setOn( true );
		ForAllDocumentScripts( true, &MyQCodeEdit::setWordWrap );
	}
	if( GetActScript() && !GetActScript()->overwriteMode() )
	{
		m_pToggleOverwrite->setOn( false );
		ForAllDocumentScripts( false, &MyQCodeEdit::setOverwriteMode );
	}
	else
	{
		m_pToggleOverwrite->setOn( true );
		ForAllDocumentScripts( true, &MyQCodeEdit::setOverwriteMode );
	}
	if( m_iSpacesForTabs==0 )
	{
		ForAllDocumentScripts( true, &MyQCodeEdit::setIndentationsUseTabs );
	}
	else
	{
		ForAllDocumentScripts( false, &MyQCodeEdit::setIndentationsUseTabs );
	}
	if( m_iShowWhitespaces==0 )
	{
		ForAllDocumentScripts( QextScintilla::WsInvisible, &MyQCodeEdit::setWhitespaceVisibility );
		ForAllDocumentScripts( false, &MyQCodeEdit::setEolVisibility );
	}
	else
	{
		ForAllDocumentScripts( QextScintilla::WsVisible, &MyQCodeEdit::setWhitespaceVisibility );
		ForAllDocumentScripts( true, &MyQCodeEdit::setEolVisibility );
	}
	if( m_iShowLineNumbers==0 )
	{
		ForAllDocumentScripts( 1, true, &MyQCodeEdit::setMarginLineNumbers );
		ForAllDocumentScripts( 1, 0, (void (MyQCodeEdit::*)(int,int))&MyQCodeEdit::setMarginWidth );
	}
	else
	{
		ForAllDocumentScripts( 1, true, &MyQCodeEdit::setMarginLineNumbers );
#ifdef ZAURUS
		ForAllDocumentScripts( 1, QString("999"), (void (MyQCodeEdit::*)(int,const QString &))&MyQCodeEdit::setMarginWidth );
#else
		ForAllDocumentScripts( 1, QString("9999"), (void (MyQCodeEdit::*)(int,const QString &))&MyQCodeEdit::setMarginWidth );
#endif
	}
	if( m_iAutoIndent==0 )
	{
		ForAllDocumentScripts( false, &MyQCodeEdit::setAutoIndent );
	}
	else
	{
		ForAllDocumentScripts( true, &MyQCodeEdit::setAutoIndent );
	}
	if( m_iCodeFolding==0 )
	{
		ForAllDocumentScripts( QextScintilla::NoFoldStyle, &MyQCodeEdit::setFolding );
	}
	else
	{
		ForAllDocumentScripts( QextScintilla::BoxedTreeFoldStyle, &MyQCodeEdit::setFolding );
	}
	ForAllDocumentScripts( (QextScintilla::EolMode)m_iEolMode, &MyQCodeEdit::setEolMode );
	ForAllDocumentScripts( (QextScintilla::EolMode)m_iEolMode, &MyQCodeEdit::convertEols );
}

void VisiScript::UpdateLanguage()
{
	for( int i=0; i<GetDocumentCount(); i++ )
	{
		MyQCodeEdit * pScript = GetScriptFor( i );

		pScript->setLexer( 0 );

		QextScintillaLexer * pLexer = 0;

		switch( m_iLanguageID )
		{
			case LANG_NONE:
				pLexer = 0;
				break;
			case LANG_MINSCRIPT:
			case LANG_CPP:
				pLexer = new QextScintillaLexerCPP();
				break;
			case LANG_CSHARP:
				pLexer = new QextScintillaLexerCSharp();
				break;
			case LANG_PYTHON:
				pLexer = new QextScintillaLexerPython();
				break;
			case LANG_JAVA:
				pLexer = new QextScintillaLexerJava();
				break;
			case LANG_JAVASCRIPT:
				pLexer = new QextScintillaLexerJavaScript();
				break;
			case LANG_RUBY:
			case LANG_PHP:
				// vorlaeufig...
				pLexer = new QextScintillaLexerCPP();
				break;
			case LANG_PERL:
				pLexer = new QextScintillaLexerPerl();
				break;
			case LANG_IDL:
				pLexer = new QextScintillaLexerIDL();
				break;
			case LANG_HTML:
				pLexer = new QextScintillaLexerHTML();
				break;
			case LANG_SQL:
				pLexer = new QextScintillaLexerSQL();
				break;
			default:
				pLexer = 0;
		}

		UpdateFont();

		pScript->setLexer( pLexer );
		if( pLexer )
		{
			pLexer->refreshProperties();
		}
		//pScript->setText( pScript->text() );

		pScript->update();
		pScript->repaint();
	}
}

void VisiScript::UpdateLanguageMenu()
{
	// deselect all languages
	for( int i=LANG_NONE; i<LANG_MAX_ID; i++ )
	{
		pLanguages->setItemChecked( i, false );
	}
	// and select the new language
	pLanguages->setItemChecked( m_iLanguageID, true );
}

void VisiScript::UpdateScriptOutputRatio()
{
	sltMaximizeOutput();
}

int VisiScript::GetHeightForOutput( int iPercent )
{
	int iValue = height()*iPercent/100;
	return iValue;
}

void VisiScript::PushLastFileName( const QString & sName )
{
	if( !ExistsLastFileName( sName ) )
	{
		m_aLastFilesList.append( new QString( sName ) );

		if( m_aLastFilesList.count()>MAX_LAST_FILES )
		{
			m_aLastFilesList.removeFirst();
		}
	}
}

bool VisiScript::ExistsLastFileName( const QString & sName )
{
	for( unsigned int i=0; i<m_aLastFilesList.count(); i++ )
	{
		if( *(m_aLastFilesList.at( i )) == sName )
		{
			return true;
		}
	}

	return false;
}

bool VisiScript::DoImportFile( const QString & sFileName )
{
	if( !sFileName.isNull() )
	{
		QString sText;
		bool bOk = ReadTextFile( sFileName.latin1(), sText );

		if( bOk )
		{
			AddDocument( -1, sFileName );

	    	GetActScript()->setText( sText );
    		GetActScript()->setEdited( FALSE );
			//updateCaption( sFileName+" #" );

			PushLastFileName( sFileName );

			return true;
		}
	}
	return false;
}

bool VisiScript::DoExportFile( const QString & sFileName )
{
	QString sText;
	if( m_bSaveOutput )
	{
		sText = GetActOutput()->text();
	}
	else
	{
		sText = GetActScript()->text();
	}

	bool bOk = WriteTextFile( sFileName.latin1(), sText.latin1() );

	if( bOk && !m_bSaveOutput )
	{
    	GetActScript()->setEdited(FALSE);
		//updateCaption( sFileName+" #" );
	}

	return bOk;
}

bool VisiScript::DoExportFile( MyQCodeEdit * pScript )
{
	QString sText = pScript->text();

	bool bOk = WriteTextFile( pScript->fileName().latin1(), sText.latin1() );

	if( bOk )
	{
    	pScript->setEdited(FALSE);
	}

	return bOk;
}

void VisiScript::SetOutputSize( int iPercentSize )
{
	for( int i=0; i<GetDocumentCount(); i++ )
	{
		GetOutputFor( i )->setMinimumSize( GetOutputFor( i )->minimumSize().width(), GetHeightForOutput( iPercentSize ) );
		GetOutputFor( i )->setMaximumSize( GetOutputFor( i )->maximumSize().width(), GetHeightForOutput( iPercentSize ) );
	}
}

void VisiScript::sltEditUndo()
{
	GetActScript()->undo();
}

void VisiScript::sltEditRedo()
{
	GetActScript()->redo();
}

void VisiScript::sltEditCut()
{
	if( GetActScript()->hasMarkedTxt() )
	{
		GetActScript()->cut();
	}
}

void VisiScript::sltEditCopy()
{
	if( GetActScript()->hasMarkedTxt() )
	{
		GetActScript()->copy();
	}
	else if( GetActOutput()->hasMarkedTxt() )
	{
		GetActOutput()->copy();
	}
}

void VisiScript::sltEditPaste()
{
	GetActScript()->paste();
}

void VisiScript::sltClearOutput()
{
	GetActOutput()->setText( "" );
}

void VisiScript::sltEditFindText()
{
#ifdef ZAURUS
	FindDlg * pDlg = new FindDlg( this, "findtextdlg", /*bModal*/TRUE );;
#else
	FindDlgImpl * pDlg = new FindDlgImpl( this, "findtextdlg", /*bModal*/TRUE );;
#endif

	pDlg->adjustSize();

	// if any text is marked, use this text for searching
	if( GetActScript()->hasMarkedTxt() )
	{
		pDlg->m_pFindText->setText( GetActScript()->markedTxt() );
	}
	else
	{
		pDlg->m_pFindText->setText( m_sFindText );
	}
	pDlg->m_pRegularExpr->setChecked( m_bRegularExpr );
	pDlg->m_pCaseSensitive->setChecked( m_bCaseSensitive );
	pDlg->m_pOnlyWholeWord->setChecked( m_bOnlyWholeWords );
	pDlg->m_pSearchBackward->setChecked( m_bSearchBackward );

	int iRet = pDlg->exec();

	if( iRet == 1 )
	{
		m_bIsReplace = false;

		m_sFindText = pDlg->m_pFindText->text();
		m_bRegularExpr = pDlg->m_pRegularExpr->isChecked();
		m_bCaseSensitive = pDlg->m_pCaseSensitive->isChecked();
		m_bOnlyWholeWords = pDlg->m_pOnlyWholeWord->isChecked();
		m_bSearchBackward = pDlg->m_pSearchBackward->isChecked();

		GetActScript()->findFirst( m_sFindText,
								   m_bRegularExpr,
		 						   m_bCaseSensitive,
								   m_bOnlyWholeWords,
								   !m_bSearchBackward );
	}

	delete pDlg;
}

void VisiScript::sltEditFindNext()
{
	bool bFound = GetActScript()->findNext();

	if( m_bIsReplace && bFound )
	{
		GetActScript()->replace( m_sReplaceText );
	}
}

void VisiScript::sltEditReplaceText()
{
#ifdef ZAURUS
	ReplaceDlg * pDlg = new ReplaceDlg( this, "replacetextdlg", /*bModal*/TRUE );;
#else
	ReplaceDlgImpl * pDlg = new ReplaceDlgImpl( this, "replacetextdlg", /*bModal*/TRUE );;
#endif

	pDlg->adjustSize();

	// if any text is marked, use this text for searching
	if( GetActScript()->hasMarkedTxt() )
	{
		pDlg->m_pFindText->setText( GetActScript()->markedTxt() );
	}
	else
	{
		pDlg->m_pFindText->setText( m_sFindText );
	}
	pDlg->m_pReplaceText->setText( m_sReplaceText );
	pDlg->m_pRegularExpr->setChecked( m_bRegularExpr );
	pDlg->m_pCaseSensitive->setChecked( m_bCaseSensitive );
	pDlg->m_pOnlyWholeWord->setChecked( m_bOnlyWholeWords );
	pDlg->m_pSearchBackward->setChecked( m_bSearchBackward );
	pDlg->m_pReplaceAll->setChecked( m_bReplaceAll );

	int iRet = pDlg->exec();

	if( iRet == 1 )
	{
		m_bIsReplace = true;

		m_sFindText = pDlg->m_pFindText->text();
		m_sReplaceText = pDlg->m_pReplaceText->text();
		m_bRegularExpr = pDlg->m_pRegularExpr->isChecked();
		m_bCaseSensitive = pDlg->m_pCaseSensitive->isChecked();
		m_bOnlyWholeWords = pDlg->m_pOnlyWholeWord->isChecked();
		m_bSearchBackward = pDlg->m_pSearchBackward->isChecked();
		m_bReplaceAll = pDlg->m_pReplaceAll->isChecked();

		bool bFound = GetActScript()->findFirst( m_sFindText,
												 m_bRegularExpr,
		 										 m_bCaseSensitive,
												 m_bOnlyWholeWords,
												 !m_bSearchBackward );

		if( bFound )
		{
			GetActScript()->beginUndoAction();

			GetActScript()->replace( m_sReplaceText );

			while( m_bReplaceAll && bFound )
			{
				bFound = GetActScript()->findNext();

				if( m_bIsReplace && bFound )
				{
					GetActScript()->replace( m_sReplaceText );
				}
			}

			GetActScript()->endUndoAction();
		}
	}

	delete pDlg;
}

void VisiScript::sltClearScript()
{
	CheckActData();

	GetActScript()->setText( "" );
    GetActScript()->setEdited(FALSE);
}

#define MAX_BUFFER_LEN 512

void VisiScript::sltRun()
{
	QString sScript = GetActScript()->text();

	if( !sScript.isNull() )
	{
		WriteTextFile( _TEMP_SCRIPT_FILE, (const char *)sScript );
	}

	char sCommand[MAX_BUFFER_LEN];
	const char * sScriptCmd = (const char *)m_sScriptCmd;

	QString sOutput;

	if( m_iExecuteMode == EXECUTE_MODE_POPEN )
	{
		sprintf( sCommand, "%s %s 2>%s", (sScriptCmd ? sScriptCmd : ""), _TEMP_SCRIPT_FILE, _TEMP_ERR_FILE );

#ifdef ZAURUS
		FILE * pPythonOutput = popen( sCommand, "r" );

		char sBuffer[MAX_BUFFER_LEN];
		size_t iSize = 0;
		do {
			iSize = fread( sBuffer, 1, MAX_BUFFER_LEN-1, pPythonOutput );
			sBuffer[iSize] = 0;
			if( iSize>0 )
			{
				sOutput += sBuffer;
			}
		} while( iSize>0 );

		pclose( pPythonOutput );
#else
		int iRet = spawnlp(_P_WAIT,"cmd","/C",sScriptCmd,_TEMP_SCRIPT_FILE,_TEMP_ERR_FILE_PIPE,_TEMP_OUT_FILE_PIPE,NULL);
#endif
	}
	else
	{
		sprintf( sCommand, "%s %s 2>%s 1>%s", (sScriptCmd ? sScriptCmd : ""), _TEMP_SCRIPT_FILE, _TEMP_ERR_FILE, _TEMP_OUT_FILE );

#ifdef ZAURUS
		int iStatus;
		pid_t pid;

		pid = fork();
		if( pid == 0 )
		{
			execl( (const char *)m_sExeclCmd, (const char *)m_sExeclCmd, "--login", "-c", sCommand, 0 );
			exit( EXIT_FAILURE );
		}
		else if( pid < 0 )
		{
			/* fork failed ! */
			iStatus = -1;
			GetActOutput()->setText( GetActOutput()->text() + "Error forking process." );
		}
		else
		{
			if( waitpid( pid, &iStatus, 0 ) != pid )
			{
				iStatus = -1;
			}
		}
#else
		int iRet = spawnlp(_P_WAIT,"cmd","/C",sScriptCmd,_TEMP_SCRIPT_FILE,_TEMP_ERR_FILE_PIPE,_TEMP_OUT_FILE_PIPE,NULL);
#endif
	}


	// show the output of the script in any case
	GetActOutput()->setText( GetActOutput()->text() + sOutput );

	// and also show the error-message, if available
	QString sErrText;
	bool bOk = ReadTextFile( _TEMP_ERR_FILE, sErrText );
	if( bOk && !sErrText.isNull() && sErrText.length()>0 )
	{
		sErrText = "Error: " + sErrText;
		GetActOutput()->setText( GetActOutput()->text() + sErrText );
	}
	remove( _TEMP_ERR_FILE );

	if( 
#ifndef ZAURUS
		// for windows allways redirecting the output into files !
		true ||
#endif
		m_iExecuteMode == EXECUTE_MODE_EXECL )
	{
		// and also show the error-message, if available
		QString sOutText;
		bool bOk2 = ReadTextFile( _TEMP_OUT_FILE, sOutText );
		if( bOk2 && !sOutText.isNull() && sOutText.length()>0 )
		{
			GetActOutput()->setText( GetActOutput()->text() + sOutText );
		}
		remove( _TEMP_OUT_FILE );
	}

	GetActOutput()->setCursorPosition( GetActOutput()->numLines(), 1 );

	remove( _TEMP_SCRIPT_FILE );
}

bool VisiScript::CheckActData()
{
	if( GetActScript()->edited() )
	{
		int iRet = QMessageBox::information( this, tr( "Save file" ), tr( "File was modified, save now ?" ), QMessageBox::Yes, QMessageBox::No );

		if( iRet == QMessageBox::Yes )
		{
			sltFileSave();
		}
	}
	return true;
}

bool VisiScript::CheckAllData()
{
	for( int i=0; i<GetDocumentCount(); i++ )
	{
		MyQCodeEdit * pScript = GetScriptFor( i );

		if( pScript->edited() )
		{
			int iRet = QMessageBox::information( this, tr( "Save file" ), tr( "File %1 was modified, save now ?" ).arg(pScript->fileName()), QMessageBox::Yes, QMessageBox::No );

			if( iRet == QMessageBox::Yes )
			{
				DoExportFile( pScript );
			}
		}
	}
	return true;
}

void VisiScript::sltNewFile()
{
    sltClearScript();

    sltShowScript();
}

void VisiScript::sltShowScript()
{
#ifdef ZAURUS
    m_pFileSelector->hide();
    m_pMenu->show();
    m_pButtonBar->show();
    m_pStack->raiseWidget( m_pTabWidget );
    GetActScript()->setFocus();
#else
    GetActScript()->setFocus();
#endif
}

void VisiScript::show()
{
	QMainWindow::show();

	SetOutputSize( m_iOutputRatio );
}

void VisiScript::sltFileAboutToShow()
{
}

void VisiScript::sltEditAboutToShow()
{
	bool bEnableCopy = GetActScript()->hasMarkedTxt() || GetActOutput()->hasMarkedTxt();
	m_pCopyAction->setEnabled( bEnableCopy );
	m_pCutAction->setEnabled( bEnableCopy );

	QClipboard * pClipboard = QApplication::clipboard();
	QString sTemp = pClipboard->text();
	bool bEnablePaste = !sTemp.isNull();
	m_pPasteAction->setEnabled( bEnablePaste );
}

void VisiScript::sltViewAboutToShow()
{
	m_pToggleOverwrite->setOn( GetActScript()->overwriteMode() );
}

void VisiScript::sltLanguagesAboutToShow()
{
	UpdateLanguageMenu();
}

void VisiScript::sltLanguageActivated( int iId )
{
	m_iLanguageID = iId;
	UpdateLanguage();
}

void VisiScript::updateCaption( const QString & sName )
{
	if( sName.isNull() )
	{
		setCaption( tr( "VisiScript" ) );
	}
	else
	{
		setCaption( sName + " - " + tr( "VisiScript" ) );
	}
}

const char * c_sSection 		= "Script";
const char * c_sScriptCmdKey 	= "ScriptCmd";
const char * c_sExtentionKey	= "Extention";
const char * c_sFontNameKey 	= "FontName";
const char * c_sFontStyleKey 	= "FontStyle";
const char * c_sFontSizeKey 	= "FontSize";
const char * c_sWordWrapKey		= "WordWrap";
const char * c_sTabSpaceKey		= "TabSpace";
const char * c_sLanguageID		= "LanguageID";
const char * c_sOverwriteKey	= "Overwrite";
const char * c_sOutputRatio		= "OutputRatio";
const char * c_sLastDirectoryKey = "LastDirectory";
const char * c_sExeclCmdKey		= "ExeclCmd";
const char * c_sExecuteModeKey	= "ExecuteMode";
const char * c_sShowStaruptFileDialog	= "ShowFileDialog";
const char * c_sSpacesForTabs = "SpacesForTabs";
const char * c_sShowWhitespaces = "ShowWhitespaces";
const char * c_sShowLineNumbers	= "ShowLineNumbers";
const char * c_sAutoIndent		= "AutoIndent";
const char * c_sCodeFolding		= "CodeFolding";
const char * c_sEolMode			= "EolMode";
const char * c_sLastFileName	= "LastFileName";

void VisiScript::loadInit()
{
#ifdef ZAURUS
	Config aCfg( _VISISCRIPT_NAME );

	aCfg.setGroup( c_sSection );
#else
	QSettings aCfg;

	aCfg.setPath("mneuroth.de","visiscript",QSettings::User);
#endif

	m_sScriptExtention = GetExtentionFromScriptCmd( m_sScriptCmd );

	m_sScriptCmd = aCfg.readEntry( c_sScriptCmdKey, m_sScriptCmd );
	m_sScriptExtention = aCfg.readEntry( c_sExtentionKey, m_sScriptExtention );
	m_sFontName = aCfg.readEntry( c_sFontNameKey, m_sFontName );
	m_sFontStyle = aCfg.readEntry( c_sFontStyleKey, m_sFontStyle );
	m_iFontSize = aCfg.readNumEntry( c_sFontSizeKey, m_iFontSize );
	m_iWordWrap = aCfg.readNumEntry( c_sWordWrapKey, m_iWordWrap );
	m_iOverwrite = aCfg.readNumEntry( c_sOverwriteKey, m_iOverwrite );
	m_iTabSpace = aCfg.readNumEntry( c_sTabSpaceKey, m_iTabSpace );
	m_iLanguageID = aCfg.readNumEntry( c_sLanguageID, m_iLanguageID );
	m_iEolMode = aCfg.readNumEntry( c_sEolMode, m_iEolMode );
	m_iOutputRatio = aCfg.readNumEntry( c_sOutputRatio, m_iOutputRatio );
	m_iSpacesForTabs = aCfg.readNumEntry( c_sSpacesForTabs, m_iSpacesForTabs );
	m_iShowWhitespaces = aCfg.readNumEntry( c_sShowWhitespaces, m_iShowWhitespaces );
	m_iShowLineNumbers = aCfg.readNumEntry( c_sShowLineNumbers, m_iShowLineNumbers );
	m_iAutoIndent = aCfg.readNumEntry( c_sAutoIndent, m_iAutoIndent );
	m_iCodeFolding = aCfg.readNumEntry( c_sCodeFolding, m_iCodeFolding );
	m_sLastDirectory = aCfg.readEntry( c_sLastDirectoryKey, m_sLastDirectory );
	m_sExeclCmd = aCfg.readEntry( c_sExeclCmdKey, m_sExeclCmd );
	m_iExecuteMode = aCfg.readNumEntry( c_sExecuteModeKey, m_iExecuteMode );
	m_iShowStartupFileDialog = aCfg.readNumEntry( c_sShowStaruptFileDialog, m_iShowStartupFileDialog );

	for( unsigned int i=0; i<MAX_LAST_FILES; i++ )
	{
		char sBuffer[512];
		QString sLastFileName;

		sprintf( sBuffer, "%s.%d", c_sLastFileName, i );
		sLastFileName = aCfg.readEntry( sBuffer, sLastFileName );

		if( !sLastFileName.isEmpty() )
		{
			PushLastFileName( sLastFileName );
		}
	}
}

void VisiScript::saveInit()
{
#ifdef ZAURUS
	Config aCfg( _VISISCRIPT_NAME );

	aCfg.setGroup( c_sSection );
#else
	QSettings aCfg;

	aCfg.setPath("mneuroth.de","visiscript",QSettings::User);
#endif

	aCfg.writeEntry( c_sScriptCmdKey, m_sScriptCmd );
	aCfg.writeEntry( c_sExtentionKey, m_sScriptExtention );
	aCfg.writeEntry( c_sFontNameKey, m_sFontName );
	aCfg.writeEntry( c_sFontStyleKey, m_sFontStyle );
	aCfg.writeEntry( c_sFontSizeKey, m_iFontSize );
	aCfg.writeEntry( c_sWordWrapKey, m_iWordWrap );
	aCfg.writeEntry( c_sOverwriteKey, (int)GetActScript()->overwriteMode()/*m_iOverwrite*/ );
	aCfg.writeEntry( c_sTabSpaceKey, m_iTabSpace );
	aCfg.writeEntry( c_sLanguageID, m_iLanguageID );
	aCfg.writeEntry( c_sEolMode, m_iEolMode );
	aCfg.writeEntry( c_sOutputRatio, m_iOutputRatio );
	aCfg.writeEntry( c_sSpacesForTabs, m_iSpacesForTabs );
	aCfg.writeEntry( c_sShowWhitespaces, m_iShowWhitespaces );
	aCfg.writeEntry( c_sShowLineNumbers, m_iShowLineNumbers );
	aCfg.writeEntry( c_sAutoIndent, m_iAutoIndent );
	aCfg.writeEntry( c_sCodeFolding, m_iCodeFolding );
	aCfg.writeEntry( c_sLastDirectoryKey, m_sLastDirectory );
	aCfg.writeEntry( c_sExeclCmdKey, m_sExeclCmd );
	aCfg.writeEntry( c_sExecuteModeKey, m_iExecuteMode );
	aCfg.writeEntry( c_sShowStaruptFileDialog, m_iShowStartupFileDialog );

	for( unsigned int i=0; i<m_aLastFilesList.count(); i++ )
	{
		char sBuffer[512];
		QString sLastFileName = *(m_aLastFilesList.at( i ));

		sprintf( sBuffer, "%s.%d", c_sLastFileName, i );
		aCfg.writeEntry( sBuffer, sLastFileName );
	}
}

void VisiScript::closeEvent( QCloseEvent * pCloseEvent )
{
	CheckAllData();

	saveInit();

	pCloseEvent->accept();
}


MyQCodeEdit * VisiScript::GetActScript() const
{
	if( GetActDocumentTabPage() )
	{
		QObject * pChild = GetActDocumentTabPage()->child( _SCRIPT_NAME );

		return (MyQCodeEdit *)pChild;
	}
	return 0;
}

MyQMultiLineEdit * VisiScript::GetActOutput() const
{
	QObject * pChild = GetActDocumentTabPage()->child( _OUTPUT_NAME );

	return (MyQMultiLineEdit *)pChild;
}

MyQCodeEdit * VisiScript::GetScriptFor( int iIndex ) const
{
	if( iIndex>=0 && iIndex<GetDocumentCount() )
	{
#ifdef ZAURUS
		QWidget * pPage = (QWidget *)m_pTabWidget->queryList("MyQCodeEdit")->at(iIndex);
		QObject * pChild = pPage;
#else
		QWidget * pPage = m_pTabWidget->page( iIndex );
		QObject * pChild = pPage->child( _SCRIPT_NAME );
#endif
		return (MyQCodeEdit *)pChild;
	}
	return 0;
}

MyQMultiLineEdit * VisiScript::GetOutputFor( int iIndex ) const
{
	if( iIndex>=0 && iIndex<GetDocumentCount() )
	{
#ifdef ZAURUS
		QWidget * pPage = (QWidget *)m_pTabWidget->queryList("MyQMultiLineEdit")->at(iIndex);
		QObject * pChild = pPage;
#else
		QWidget * pPage = m_pTabWidget->page( iIndex );
		QObject * pChild = pPage->child( _OUTPUT_NAME );
#endif
		return (MyQMultiLineEdit *)pChild;
	}
	return 0;
}

QWidget * VisiScript::GetActDocumentTabPage() const
{
	return m_pTabWidget->currentPage();
}

int VisiScript::GetDocumentCount() const
{
#ifdef ZAURUS
	return m_pTabWidget->queryList("MyQCodeEdit")->count();
#else
	return m_pTabWidget->count();
#endif
}

void VisiScript::AddDocument( int iDocCount, const QString & sFileName )
{
	QString sFileNameTemp = sFileName;

	if( iDocCount>=0 )
	{
		sFileNameTemp = sFileNameTemp.sprintf( "%s%d", (const char *)sFileName, iDocCount );
	}

    QWidget * pTab = new QWidget( m_pTabWidget, "pTab" );
    m_pTabWidget->addTab( pTab, sFileNameTemp );

	QVBoxLayout * pLayout =  new QVBoxLayout( pTab );

    QVBox * pVBox = new QVBox( /*m_pStack*/pTab );
	pVBox->setSpacing( 4 );
	pVBox->setMargin( 2 );

	pLayout->addWidget( pVBox );

    MyQCodeEdit * pScript = new MyQCodeEdit( pVBox, _SCRIPT_NAME, m_pTabWidget, pTab, this );
	pScript->setTabWidth( m_iTabSpace );
	pScript->setLexer( 0 );
	pScript->setMarginLineNumbers( 1, true );
	pScript->setOverwriteMode( (bool)m_iOverwrite );
	pScript->setSelectionBackgroundColor( QColor( 0,0,0 ) );
	pScript->setSelectionForegroundColor( QColor( 255,255,255 ) );
	pScript->setFileName( sFileNameTemp );
	//pScript->setFolding( QextScintilla::BoxedTreeFoldStyle );

    MyQMultiLineEdit * pOutput = new MyQMultiLineEdit( pVBox, _OUTPUT_NAME );
    pOutput->setReadOnly( TRUE );

	// ** script edit field should be larger than the output edit field
    pVBox->setStretchFactor( pScript, 3 );
	//sltMaximizeOutput();

	// font settings updaten
	UpdateFont();
	UpdateFlags();
	UpdateLanguage();
	//m_pFixedAction->setOn( true );
	//m_pMediumAction->setOn( true );

	SetOutputSize( m_iOutputRatio );

	m_pTabWidget->showPage( pTab );

#ifdef ZAURUS
	QMultiLineEdit::setDefaultTabStop( m_iTabSpace );
#else
	pOutput->setTabStopWidth( m_iTabSpace );
#endif
}

void VisiScript::sltCloseActDocument()
{
	// at least one document must be open !
	if( GetDocumentCount()>1 )
	{
		CheckActData();
	
		QWidget * pPage = m_pTabWidget->currentPage();
		m_pTabWidget->removePage( pPage );
		delete pPage;
	}
}

void VisiScript::sltClearActDocument()
{
	//CheckActData();	// --> will be done in sltNewFile() / sltClearScript()

	sltNewFile();
}

template <class T1>
void VisiScript::ForAllDocumentScripts( T1 pFcn )
{
	for( int i=0; i<GetDocumentCount(); i++ )
	{
		(GetScriptFor( i )->*pFcn)();
	}
}

template <class T1, class T2>
void VisiScript::ForAllDocumentScripts( T1 aValue, T2 pFcn )
{
	for( int i=0; i<GetDocumentCount(); i++ )
	{
		(GetScriptFor( i )->*pFcn)( aValue );
	}
}

template <class T1, class T2, class T3>
void VisiScript::ForAllDocumentScripts( T1 aValue1, T2 aValue2, T3 pFcn )
{
	for( int i=0; i<GetDocumentCount(); i++ )
	{
		(GetScriptFor( i )->*pFcn)( aValue1, aValue2 );
	}
}

template <class T1>
void VisiScript::ForAllDocumentOutput( T1 pFcn )
{
	for( int i=0; i<GetDocumentCount(); i++ )
	{
		(GetOutputFor( i )->*pFcn)();
	}
}

template <class T1, class T2>
void VisiScript::ForAllDocumentOutput( T1 aValue, T2 pFcn )
{
	for( int i=0; i<GetDocumentCount(); i++ )
	{
		(GetOutputFor( i )->*pFcn)( aValue );
	}
}

#ifndef ZAURUS

#include <qdragobject.h> 

static bool IsTextFileDrag( const QMimeSource * pEvent, QString & sFileNameOut )
{
	QStrList aStrList;

	if( QUriDrag::decode( pEvent, aStrList ) )
	{
		if( aStrList.count() == 1 )
		{
			const QString sTest = aStrList.at( 0 );
			const char * s = (const char *)sTest;
			sFileNameOut = QUriDrag::uriToLocalFile( s );
			s = (const char *)sFileNameOut;

			return true;
		}
	}

	return false;
}

void VisiScript::dragEnterEvent( QDragEnterEvent * pEvent )
{
	pEvent->accept( QUriDrag::canDecode( pEvent ) );		
}

void VisiScript::dropEvent( QDropEvent * pEvent )
{
	QString sFileName;

	if( IsTextFileDrag( pEvent, sFileName ) )
	{
		DoImportFile( sFileName );
	}
}

#endif
